#include "Collision.h"

/**
 * Construtor
 * @param graph Grafo do grid
 */
Collision::Collision(Graph* graph) : graph_(graph) {
    ret = new Rect(0,0, graph_->getRows()*TILE_WIDTH, graph_->getColumns()*TILE_HEIGHT);
    qt = new Quadtree(0, ret);
}

/**
 * Destrutor
 */
Collision::~Collision() {
    printf("Deletando ret e qt de Collision\n");
    delete ret;
    delete qt;
}

/**
 * @return Ponteiro da Quadtree
 */
Quadtree* Collision::getQt() {
    return qt;
}

/**
 * Verifica se o ponto (px,py) e' andavel
 * @param px Coordenada x em pixel
 * @param py Coordenada y em pixel
 * @return true se o ponto (px,py) for andavel
 */
bool Collision::isWalkable(float px, float py) {
    return graph_->getNode()[mymath::div(px,TILE_WIDTH)][mymath::div(py,TILE_HEIGHT)]->isWalkable();
}

/**
 * Calcula o vetor que aponta o centro da unidade A ao centro da unidade B
 * @param unitA Ponteiro da unidade A
 * @param unitB Ponteiro da unidade B
 * @return Vetor (dx,dy) calculado
 */
std::pair<float,float> Collision::getCollision(Unit* unitA, Unit* unitB) {
    float e;
    float ex = 0;
    float ey = 0;
    if((unitA != NULL) && (unitB != NULL) && (unitA != unitB)) {
        float d = mymath::distance(unitA->getPx(), unitA->getPy(), unitB->getPx(), unitB->getPy());
        float dx = unitB->getPx() - unitA->getPx();
        float dy = unitB->getPy() - unitA->getPy();
        float cosseno = 0;
        float seno = 0;
        // Se d == 0 então unitA está exatamente no mesmo ponto de unitB e tbm não queremos fazer divisão por 0
        if(d != 0) {
            cosseno = dx/d;
            seno = dy/d;
        }
        e = unitA->getRadius()+unitB->getRadius()-d; // O quanto colidiu
        // Se for <= nao estao colididos
        if(e > 0) {
            ex = cosseno*e;
            ey = seno*e;
        }
    }
    return std::make_pair(ex,ey);
}

/**
 * Resolve colisao para um conjunto de unidades
 * @param unitSet Vetor de unidades
 */
void Collision::solve(std::vector<Unit*> unitSet) {
    Unit* unitA;
    Unit* unitB;
    std::vector<Unit*> returnObjects;
    std::vector<std::pair<float,float>> collisionPath;
    qt->clear();
    // Insere unidades na Quadtree
    for (int i = 0; i < (int) unitSet.size(); i++) {
        qt->insert(unitSet[i]);
    }
    for(int i = 0; i < (int) unitSet.size(); i++) {
        unitA = unitSet[i];
        // Se a unidade A esta dando a volta em alguem
        if(unitA->inCollision()) {
            if(!isWalkable(unitA->getPx(),unitA->getPy())) {
                unitA->setCoordinates(unitA->getPx()-unitA->getLastMove().first, unitA->getPy()-unitA->getLastMove().second);
                //printf("signal: %d\n", unitA->getSignal());
                unitA->setSignal(-unitA->getSignal());
                std::vector<std::pair<float,float>> cp;
                unitA->setCollisionPath(cp);
                //unitA->stop();
                //break;
            }
        }
        returnObjects.clear();
        // Pega conjunto de unidades que esta no mesmo quadrante da unidade A
        returnObjects = qt->retrieve(returnObjects, unitA);
        // Compara somente as unidades que estao no mesmo quadrando da unidade A
        for(int j = 0; j < (int) returnObjects.size(); j++) {
            unitB = returnObjects[j];
            if(unitA != unitB) {
                float pxa = unitA->getPx();
                float pya = unitA->getPy();
                float ra = unitA->getRadius();
                float pxb = unitB->getPx();
                float pyb = unitB->getPy();
                float rb = unitB->getRadius();
                collisionPath.clear();
                std::pair<float,float> e = getCollision(unitA, unitB);
                float lastxa = 0;
                float lastya = 0;
                float lastxb = 0;
                float lastyb = 0;
                // Se A colidiu com B
                if(mymath::collide(pxa, pya, ra, pxb, pyb, rb)) {
                    //printf("A colidiu com B: e = (%f,%f)\n", e.first, e.second);
                    float distanceA = unitA->getPixelPathDistance();
                    float distanceB = unitB->getPixelPathDistance();
                    if(unitA->inCollision() && unitB->inCollision()) {
                        if(distanceA <= distanceB) {
                            unitB->setSignal(unitA->getSignal());
                            unitB->setPixelPath(unitA->getPixelPath());
                            unitB->ncollide = 0;
                        } else {
                            unitA->setSignal(unitB->getSignal());
                            unitA->setPixelPath(unitB->getPixelPath());
                            unitA->ncollide = 0;
                        }
                    }
                    // Sistema de dar a volta nas unidades que estao em estado hold position
                    if(!unitA->isHold() && unitB->isHold()) {
                        int signal;
                        unitA->setCoordinates(pxa-e.first, pya-e.second);
                        // Se a unidade A nao estiver dando a volta em alguma outra
                        if(unitA->collide == NULL) {
                            //unitA->setCoordinates(pxa+e.second-e.first, pya-e.first-e.second); // Roda pra um lado
                            //unitA->setCoordinates(pxa-e.second-e.first, pya+e.first-e.second); // Roda pro outro lado
                            pxa = unitA->getPx();
                            pya = unitA->getPy();
                            float temptx = -1;
                            float tempty = -1;
                            if(unitA->getFollow() != NULL) {
                                temptx = unitA->getFollow()->getPx();
                                tempty = unitA->getFollow()->getPy();

                            } else {
                                temptx = unitA->getTx();
                                tempty = unitA->getTy();
                            }
                            if(mymath::distance(pxb+e.second,pyb-e.first, temptx, tempty) < mymath::distance(pxb-e.second,pyb+e.first, temptx, tempty)) signal = 1;
                            else signal = -1;

                            unitA->setSignal(signal);
                            for(int i = 0; i <= 3; i++) {
                                float angulo = signal*i*(M_PI/6) + atan2(pyb-pya, pxb-pxa) + M_PI;
                                float voltax = pxb+(ra+rb+1)*cos(angulo);
                                float voltay = pyb+(ra+rb+1)*sin(angulo);
                                collisionPath.push_back(std::make_pair(voltax, voltay));
                                //printf("Inseriu: %f %f - Angulo = %f %f\n", voltax, voltay, cos(angulo), sin(angulo));
                            }
                            unitA->setCollisionPath(collisionPath);
                            unitA->collide = unitB;
                            unitA->dist = mymath::distance(pxa,pya, temptx,tempty);
                        }
                        // Se a unidade A estiver dando a volta em alguma outra, faz um tratamento para ver se troca o sentido da rotacao
                        else {
                            signal = unitA->getSignal();
                            //printf("signal antes: %d\n", signal);
                            float temptx = -1;
                            float tempty = -1;
                            if(unitA->getFollow() != NULL) {
                                temptx = unitA->getFollow()->getPx();
                                tempty = unitA->getFollow()->getPy();

                            } else {
                                temptx = unitA->getTx();
                                tempty = unitA->getTy();
                            }
                            //printf("unitA->dist: %f, (A-t)->dist: %f\n", unitA->dist, mymath::distance(pxa,pya, temptx,tempty));
                            if(unitA->dist*1.25 < mymath::distance(pxa,pya, temptx,tempty) && unitA->ncollide >= 5) {
                                signal = -signal;
                                unitA->ncollide = 0;
                            }
                            //printf("signal depois: %d\n", signal);

                            for(int i = 0; i <= 3; i++) {
                                float angulo = signal*i*(M_PI/6) + atan2(pyb-pya, pxb-pxa) + M_PI;
                                float voltax = pxb+(ra+rb+1)*cos(angulo);
                                float voltay = pyb+(ra+rb+1)*sin(angulo);
                                collisionPath.push_back(std::make_pair(voltax, voltay));
                                //printf("Inseriu: %f %f - Angulo = %f %f\n", voltax, voltay, cos(angulo), sin(angulo));
                            }
                            unitA->setSignal(signal);
                            unitA->setCollisionPath(collisionPath);
                            unitA->collide = unitB;
                            unitA->ncollide++;
                        }

                    }
                    // Se a unidade B estiver no modo Hold
                    if(unitA->isHold() && !unitB->isHold()) continue;
                    // Se a unidade A e B sao de times diferentes
                    if(unitA->getTeam() != unitB->getTeam()) {
                        unitA->setCoordinates(unitA->getPx()-e.first/2, unitA->getPy()-e.second/2);
                        unitB->setCoordinates(unitB->getPx()+e.first/2, unitB->getPy()+e.second/2);
                        continue;
                    }
                    // Sistema de empurrar unidades
                    if((!unitA->isHold() && !unitB->isHold()) || (unitA->isHold() && unitB->isHold())) {
                        // Unidade A movendo e B parada
                        if(unitA->isMoving() && !unitB->isMoving()) {
                            if(isWalkable(pxb+e.first, pyb)) {
                                lastxb = e.first;
                            } else if(isWalkable(pxa-e.first, pya)) {
                                lastxa = -e.first;
                            }
                            if(isWalkable(pxb, pyb+e.second)) {
                                lastyb = e.second;
                            } else if(isWalkable(pxa, pya-e.second)) {
                                lastya = -e.second;
                            }
                        } else if(!unitA->isMoving() && unitB->isMoving()) {
                            if(isWalkable(pxa-e.first, pya)) {
                                lastxa = -e.first;
                            } else if(isWalkable(pxb+e.first, pyb)) {
                                lastxb = e.first;
                            }
                            if(isWalkable(pxa, pya-e.second)) {
                                lastya = -e.second;
                            } else if(isWalkable(pxb, pyb+e.second)) {
                                lastyb = e.second;
                            }

                        }
                        // Unidade A parada e B parada
                        else if(!unitA->isMoving() && !unitB->isMoving()) {
                            if(isWalkable(pxa-e.first/2, pya) && isWalkable(pxb+e.first/2, pyb)) {
                                lastxa = -e.first/2;
                                lastxb = e.first/2;
                            } else if(isWalkable(pxa-e.first, pya)) {
                                lastxa = -e.first;
                            } else if(isWalkable(pxb+e.first, pyb)) {
                                lastxb = e.first;
                            }
                            if(isWalkable(pxa, pya-e.second/2) && isWalkable(pxb, pyb+e.second/2)) {
                                lastya = -e.second/2;
                                lastyb = e.second/2;
                            } else if(isWalkable(pxa, pya-e.second)) {
                                lastya = -e.second;
                            } else if(isWalkable(pxb, pyb+e.second)) {
                                lastyb = e.second;
                            }
                        }
                        // Unidade A movendo e B movendo
                        else if(unitA->isMoving() && unitB->isMoving()) {
                            if(distanceA < distanceB) {
                                if(isWalkable(pxb+e.first, pyb)) {
                                    lastxb = e.first;
                                } else if(isWalkable(pxa-e.first, pya)) {
                                    lastxa = -e.first;
                                }
                                if(isWalkable(pxb, pyb+e.second)) {
                                    lastyb = e.second;
                                } else if(isWalkable(pxa, pya-e.second)) {
                                    lastya = -e.second;
                                }
                            }
                            else if(distanceA >= distanceB) {
                                if(isWalkable(pxa-e.first, pya)) {
                                    lastxa = -e.first;
                                } else if(isWalkable(pxb+e.first, pyb)) {
                                    lastxb = e.first;
                                }
                                if(isWalkable(pxa, pya-e.second)) {
                                    lastya = -e.second;
                                } else if(isWalkable(pxb, pyb+e.second)) {
                                    lastyb = e.second;
                                }
                            }

                        }
                    }
                    //printf("lastA: (%f,%f), lastB: (%f,%f)\n", lastxa, lastya, lastxb, lastyb);
                    // Se a condicao abaixo for verdadeira, nao havera mais colisao entre as duas unidades
                    if(fabs(lastxb-lastxa) == fabs(e.first) && fabs(lastyb-lastya) == fabs(e.second)) {
                        unitA->setCoordinates(pxa+lastxa, pya+lastya);
                        unitB->setCoordinates(pxb+lastxb, pyb+lastyb);
                        unitA->setLastMove(std::make_pair(lastxa,lastya));
                        unitB->setLastMove(std::make_pair(lastyb,lastyb));
                        // Depois que foi empurrado verifica se foi empurrado pra algum ponto proximo de algum outro path
                        if(unitA->isMoving()) unitA->checkPath();
                        if(unitB->isMoving()) unitB->checkPath();
                    }
                    // Ocorre quando duas unidades estao colididas e ambas nao podem ser separadas pois serao empurradas para parede
                    else {
                        /*
                        float evalue = sqrt(e.first*e.first+e.second*e.second);
                        unitA->setCoordinates(pxa+evalue*(e.first/fabs(e.first)),pya+evalue*(e.second/fabs(e.second)));
                        unitB->setCoordinates(pxb+evalue*(e.first/fabs(e.first)),pyb+evalue*(e.second/fabs(e.second)));
                        */
                    }
                }
            }
        }
    }
}
