#include "Pathfinding.h"

/**
 * Construtor
 * @param graph Grafo do grid
 */
Pathfinding::Pathfinding(Graph* graph) : graph_(graph) {
}

/**
 * Destrutor
 */
Pathfinding::~Pathfinding() {
}

/**
 * Busca caminho do ponto (sx,sy) ao ponto (tx,ty) usando o algoritmo BFS
 * @param sx Coordenada x em pixel do ponto origem
 * @param sy Coordenada y em pixel do ponto origem
 * @param tx Coordenada x em pixel do ponto destino
 * @param ty Coordenada y em pixel do ponto destino
 * @return Caminho encontrado
 */
std::vector<std::pair<float,float>> Pathfinding::bfsPathfinding(float sx, float sy, float tx, float ty) {
    Node*** node = graph_->getNode();
    Node* s = node[mymath::div(sx,TILE_WIDTH)][mymath::div(sy,TILE_HEIGHT)];
    Node* t = node[mymath::div(tx,TILE_WIDTH)][mymath::div(ty,TILE_HEIGHT)];
    std::vector<std::pair<int,int>> path;
    std::vector<std::pair<float,float>> pixelPath;
    if(graph_->bfs(s, t)) {
        path = getPath(s,t);
        //path.erase(path.begin()); // Apaga a posicao inicial da unidade dentro path
        //pixelPath.push_back(std::make_pair(tx, ty));
        if(path.size() == 0) return pixelPath;
        return transformPath(path, tx, ty);
    }
    return pixelPath;
}

/**
 * Busca caminho do ponto (sx,sy) ao ponto (tx,ty) usando o algoritmo A*
 * @param sx Coordenada x em pixel do ponto origem
 * @param sy Coordenada y em pixel do ponto origem
 * @param tx Coordenada x em pixel do ponto destino
 * @param ty Coordenada y em pixel do ponto destino
 * @return Caminho encontrado
 */
std::vector<std::pair<float,float>> Pathfinding::astarPathfinding(float sx, float sy, float tx, float ty) {
    Node*** node = graph_->getNode();
    Node* s = node[mymath::div(sx,TILE_WIDTH)][mymath::div(sy,TILE_HEIGHT)];
    Node* t = node[mymath::div(tx,TILE_WIDTH)][mymath::div(ty,TILE_HEIGHT)];
    std::vector<std::pair<int,int>> path;
    std::vector<std::pair<float,float>> pixelPath;
    if(graph_->astar(s, t)) {
        path = getPath(s,t);
        //path.erase(path.begin()); // Apaga a posicao inicial da unidade dentro path
        //pixelPath.push_back(std::make_pair(tx, ty));
        //if(path.size() == 0) return pixelPath;
        return transformPath(path, tx, ty);
    }
    return pixelPath;
}

/**
 * Busca caminho do ponto (sx,sy) ao ponto (tx,ty) usando o algoritmo Dijkstra
 * @param sx Coordenada x em pixel do ponto origem
 * @param sy Coordenada y em pixel do ponto origem
 * @param tx Coordenada x em pixel do ponto destino
 * @param ty Coordenada y em pixel do ponto destino
 * @return Caminho encontrado
 */
std::vector<std::pair<float,float>> Pathfinding::dijkstraPathfinding(float sx, float sy, float tx, float ty) {
    Node*** node = graph_->getNode();
    Node* s = node[mymath::div(sx,TILE_WIDTH)][mymath::div(sy,TILE_HEIGHT)];
    Node* t = node[mymath::div(tx,TILE_WIDTH)][mymath::div(ty,TILE_HEIGHT)];
    std::vector<std::pair<int,int>> path;
    std::vector<std::pair<float,float>> pixelPath;
    if(graph_->dijkstra(s, t)) {
        path = getPath(s,t);
        //path.erase(path.begin()); // Apaga a posicao inicial da unidade dentro path
        //pixelPath.push_back(std::make_pair(tx, ty));
        if(path.size() == 0) return pixelPath;
        return transformPath(path, tx, ty);
    }

    return pixelPath;
}

/**
 * Usa o grafo do grid para obter o caminho do no' s ao no' t
 * @param s
 * @param t
 * @return Caminho do no' s ao no' t
 */
std::vector<std::pair<int,int>> Pathfinding::getPath(Node* s, Node* t) {
    std::vector<std::pair<int,int>> path;
    for(Node* current = t; ; current = current->getParent()) {
        if(current == s) {
            break;
        }
        path.push_back(std::make_pair(current->getX(), current->getY()));
        current->setOnPath(true);
    }
    std::reverse(path.begin(), path.end());
    return path;
}

/**
 * Transforma o caminho dos tiles em pixel
 * @param path Caminho
 * @param tx Coordenada x do ponto destino
 * @param ty Coordenada y do ponto destino
 * @return Caminho path transformados em posicoes pixel
 */
std::vector<std::pair<float,float>> Pathfinding::transformPath(std::vector<std::pair<int,int>> path, float tx, float ty) {
    std::vector<std::pair<float,float>> pixelPath;
    // Simplemente escolhe o meio do tile pra ir
    for(int i = 0; i < (int) path.size()-1; i++) {
        pixelPath.push_back(std::make_pair(path[i].first*TILE_WIDTH+TILE_WIDTH/2, path[i].second*TILE_HEIGHT+TILE_HEIGHT/2));
    }
    // O ultimo path é exatamente onde foi clicado
    pixelPath.push_back(std::make_pair(tx, ty));
    return pixelPath;
}

/**
 * Remove os caminhos onde walking == false.
 * Usado no buildingPath(Quando a construção foi requisitada para mover a qualquer lugar ou em cima de uma outra construção).
 * Usado no unitMove(Quando a unidade foi requisitada para mover em um ponto em cima de uma construção)
 * @param pixelPath Caminho em pixel
 * @return Caminho em pixel com remocoes
 */
std::vector<std::pair<float,float>> Pathfinding::getWalkingPath(std::vector<std::pair<float,float>> pixelPath) {
    Node*** node = graph_->getNode();
    // Remove os pixelpaths do inicio, serve somente quando uma construção faz um pathfinding
    while(pixelPath.size() > 0 && !node[mymath::div(pixelPath[0].first, TILE_WIDTH)][mymath::div(pixelPath[0].second, TILE_HEIGHT)]->isWalkable()) {
        pixelPath.erase(pixelPath.begin());
    }
    // Remove os pixelpaths do fim, serve somente quando onde foi clicado havia uma construção
    while(pixelPath.size() > 0 && !node[mymath::div(pixelPath[pixelPath.size()-1].first, TILE_WIDTH)][mymath::div(pixelPath[pixelPath.size()-1].second, TILE_HEIGHT)]->isWalkable()) {
        pixelPath.pop_back();
    }
    return pixelPath;
}

/**
 * Essa funcao devolve o conjunto apenas para nao precisar fazer um tratamento diferente em relacao aos edificios
 * @param unit Ponteiro da unidade
 * @return Conjunto de zonas de equivalencia da unidade
 */
std::set<int> Pathfinding::getZE(Unit* unit) {
    Node*** node = graph_->getNode();
    std::map<int,int> ZE = graph_->getZoneEquivalency();
    std::set<int> zones;
    zones.insert(ZE[node[mymath::div(unit->getPx(), TILE_WIDTH)][mymath::div(unit->getPy(), TILE_HEIGHT)]->getZone()]);
    return zones;
}

/**
 * @param building Ponteiro da unidade
 * @return Conjunto de zonas de equivalencia do edificio
 */
std::set<int> Pathfinding::getZE(Building* building) {
    Node*** node = graph_->getNode();
    std::map<int,int> ZE = graph_->getZoneEquivalency();
    std::set<int> zones;
    int bx = building->getX();
    int by = building->getY();
    int dim = building->getDimension();
    // Da pra fazer mais eficiente, aqui está percorrendo toda area de building
    for(int i = bx-1; i < bx+dim+1; i++) {
        for(int j = by-1; j < by+dim+1; j++) {
            if(i >= 0 && j >= 0 && i < graph_->getRows() && j < graph_->getColumns()) {
                if(ZE[node[i][j]->getZone()] != 0) {
                    zones.insert(ZE[node[i][j]->getZone()]);
                }
            }
        }
    }
    return zones;
}

/**
 * @param px Coordenada x em pixel
 * @param py Coordenada y em pixel
 * @return Zona de equivalencia do ponto (px,py)
 */
std::set<int> Pathfinding::getZE(float px, float py) {
    Node*** node = graph_->getNode();
    std::map<int,int> ZE = graph_->getZoneEquivalency();
    std::set<int> zones;
    zones.insert(ZE[node[mymath::div(px, TILE_WIDTH)][mymath::div(py, TILE_HEIGHT)]->getZone()]);
    return zones;
}

/**
 * Compara os elementos de dois conjuntos de zonas de equivalencia
 * @param ze1 Conjunto 1 de zonas de equivalencia
 * @param ze2 Conjunto 2 de zonas de equivalencia
 * @return true se algum elemento do conjunto 1 for igual a algum elemento do conjunto 2
 */
bool Pathfinding::isInSameZE(std::set<int> ze1, std::set<int> ze2) {
    /* Da pra melhorar essa comparacao */
    for(std::set<int>::iterator it1 = ze1.begin(); it1 != ze1.end(); ++it1) {
        for(std::set<int>::iterator it2 = ze2.begin(); it2 != ze2.end(); ++it2) {
            if((*it1) == (*it2)) return true;
        }
    }
    return false;
}

/**
 * Calcula o ponto (px,py) mais proximo de (tx,ty) que possui mesma ZE de alguma zona do conjunto zones.
 * Serve para quando o jogador faz uma requisicao de um pathfinding para um local inalcancavel
 * @param tx Coordenada x em pixel
 * @param ty Coordenada y em pixel
 * @param zones Conjunto de zonas de equivalencia
 * @return Ponto (px,py) encontrado
 */
std::pair<float,float> Pathfinding::findNearestTile(float tx, float ty, std::set<int> zones) {
    Node* newTarget;
    Node*** node = graph_->getNode();
    int posX = mymath::div(tx,TILE_WIDTH);
    int posY = mymath::div(ty,TILE_HEIGHT);
    Node* t = node[posX][posY];
    newTarget = graph_->noPathDijkstra(t, zones);
    return std::make_pair(newTarget->getX()*TILE_WIDTH+TILE_WIDTH/2, newTarget->getY()*TILE_HEIGHT+TILE_HEIGHT/2);
}
