/* Codigo baseado do site:
http://gamedev.tutsplus.com/tutorials/implementation/quick-tip-use-quadtrees-to-detect-likely-collisions-in-2d-space/ */

#include "Quadtree.h"

/**
 * Construtor
 * @param pLevel Nivel da pronfundidade
 * @param pBounds Retangulo da Quadtree filha
 */
Quadtree::Quadtree(int pLevel, Rect* pBounds) {
    MAX_OBJECTS = 5;
    MAX_LEVELS = 5;
    level = pLevel;
    bounds = pBounds;
    nodes = new Quadtree*[4];
    for(int i = 0; i < 4; i++) {
        nodes[i] = NULL;
    }
}

/**
 * Destrutor
 */
Quadtree::~Quadtree() {
    for(int i = 0; i < 4; i++) {
        delete nodes[i];
    }
    delete [] nodes;
}

/**
 * @return Retangulo da Quadtree
 */
Rect* Quadtree::getBounds() {
    return bounds;
}

/**
 * @return No's filhos da Quadtree
 */
Quadtree** Quadtree::getNodes() {
    return nodes;
}

/**
 * @return Vetor de objetos dentro da Quadtree
 */
std::vector<Unit*> Quadtree::getObjects() {
    return objects;
}

/**
 * @param unit Ponteiro da unidade
 * Insere unidade na Quadtree
 */
void Quadtree::insert(Unit* unit) {
    if (nodes[0] != NULL) {
        int index = getIndex(unit);
        if (index != -1) {
            nodes[index]->insert(unit);
            return;
        }
    }
    objects.push_back(unit);
    if ((int) objects.size() > MAX_OBJECTS && level < MAX_LEVELS) {
        if (nodes[0] == NULL) {
            split();
        }
        int i = 0;
        while(i < (int) objects.size()) {
            Unit* u = objects[i];
            int index = getIndex(u);
            /* Se a unidade cabe estritamente dentro do retangulo, entao remove do retangulo atual e insere no filho */
            if (index != -1) {
                nodes[index]->insert(u);
                objects.erase(objects.begin()+i);
            } else {
                i++;
            }
        }
    }
}

/**
 * Divide Quadtree em 4
 */
void Quadtree::split() {
    int subWidth = (int) (bounds->getWidth()/2);
    int subHeight = (int) (bounds->getHeight()/2);
    int x = (int) (bounds->getX());
    int y = (int) (bounds->getY());
    nodes[0] = new Quadtree(level+1, new Rect(x + subWidth, y, subWidth, subHeight));
    nodes[1] = new Quadtree(level+1, new Rect(x, y, subWidth, subHeight));
    nodes[2] = new Quadtree(level+1, new Rect(x, y + subHeight, subWidth, subHeight));
    nodes[3] = new Quadtree(level+1, new Rect(x + subWidth, y + subHeight, subWidth, subHeight));
}

/**
 * Limpa Quadtree
 */
void Quadtree::clear() {
    objects.clear();
    for(int i = 0; i < 4; i++) {
        if(nodes[i] != NULL) {
            nodes[i]->clear();
            nodes[i] = NULL;
        }
    }
}

/**
 * Devolve o indice da Quadtree em que a unidade unit pertence
 * @param unit Ponteiro da unidade
 * @return Indice da Quadtree
 */
int Quadtree::getIndex(Unit* unit) {
    int index = -1;
    double verticalMidpoint = bounds->getX() + (bounds->getWidth() / 2);
    double horizontalMidpoint = bounds->getY() + (bounds->getHeight() / 2);
    bool topQuadrant = (unit->getPy() + unit->getRadius() < horizontalMidpoint);
    bool bottomQuadrant = (unit->getPy() - unit->getRadius() > horizontalMidpoint);
    if (unit->getPx() + unit->getRadius() < verticalMidpoint) {
        if (topQuadrant) {
            index = 1;
        } else if (bottomQuadrant) {
            index = 2;
        }
    }
    else if (unit->getPx() - unit->getRadius() > verticalMidpoint) {
        if (topQuadrant) {
            index = 0;
        } else if (bottomQuadrant) {
            index = 3;
        }
    }
    return index;
}

/**
 * Devolve o conjunto de unidades que estão na mesma Quadtress de unit
 * @param returnObjects Vetor de objetos dentro da Quadtree
 * @param unit Ponteiro da unidade
 * @return Conjunto de unidades na mesma Quadtree de unit
 */
std::vector<Unit*> Quadtree::retrieve(std::vector<Unit*> returnObjects, Unit* unit) {
    int index = getIndex(unit);
    //printf("Index: %d, Level: %d\n", index, level);
    if (index != -1 && nodes[0] != NULL) {
        returnObjects = nodes[index]->retrieve(returnObjects, unit);
    }
    else if(index == -1 && nodes[0] != NULL) {
        for(int i = 0; i < 4; i++) {
            //std::vector<Unit*> obj = nodes[i]->getObjects;
            returnObjects = nodes[i]->retrieve(returnObjects, unit);
        }
    }
    for(int i = 0; i < (int) objects.size(); i++) {
        returnObjects.push_back(objects[i]);
    }
    return returnObjects;
}
