#include "PlayingState.h"
#include "PauseState.h"
#include "ResultState.h"

PlayingState* PlayingState::instance = NULL;
bool debug[6] = {false,false,false,false,false,false};
enum KEY {F1,F2,F3,F4,F5,F6};

bool unitTypeSort(Unit* a, Unit* b) {
    return a->getType() < b->getType();
}

/**
 * Construtor
 * @param display Tela do jogo
 * @param timer Tempo de atualizacao
 * @param queue Fila de eventos
 */
PlayingState::PlayingState(ALLEGRO_DISPLAY* display, ALLEGRO_TIMER* timer, ALLEGRO_EVENT_QUEUE* queue) :
    State(display, timer, queue) {
    init();
}

/**
 * Destrutor
 */
PlayingState::~PlayingState() {
    printf("Deletando collision, theme, gui, ai, game, map de PlayingState\n");
    delete collision;
    delete theme;
    delete gui;
    for(int i = 0; i < NUM_AI; i++) {
        delete ai[i];
    }
    delete ai;
    delete game_;
    delete map_;
}

/**
 * Inicializa variaveis
 */
void PlayingState::init() {
    redraw = false;
    font = al_load_ttf_font("fonts/times.ttf",20,0);
    font_gui = al_load_ttf_font("fonts/times.ttf",18,0);
    if(!font) printf("nao carregou fonte\n");
    if(!font_gui) printf("nao carregou fonte gui\n");
    fps = MAX_FPS;
    UnitProperty::getInstance()->fps = MAX_FPS; // Precisa fazer variar o fps dentro do UnitProperty
    fps_accum = MAX_FPS;
    fps_time = -1;
    zoom = 1;
    mouse = 0;
    mousex = 0;
    mousey = 0;
    scroll_x = 0;
    scroll_y = 0;
    rotate = 0;
    selectSX = selectSY = selectFX = selectFY = 0;
    theme = new Theme(font_gui);
    gui = new Gui(*theme,display_);
    gui->setRectangle(0,450,al_get_display_width(display_), al_get_display_height(display_)-450);

    action_state.push(ACTION_STATE::IDLE);
    selectBuilding = NULL;
    buildingID = -1;
    map_ = new Map("maps/map1.txt", "maps/config1.txt");
    setGame(new Game(map_));
}

/**
 * Reinicia jogo
 */
void PlayingState::restartGame() {
    delete map_;
    delete game_;
    init();
}

/**
 * Atribui game ao estado playing
 * @param game Jogo a ser atribuido
 */
void PlayingState::setGame(Game* game) {
    game_ = game;
    game_->getGraph()->setZone();
    gui->setGame(game);
    collision = new Collision(game_->getGraph());
    ai = new AI*[NUM_AI];
    for(int i = 0; i < NUM_AI; i++) {
        ai[i] = new AI(game_, game_->getPlayer()[i]);
    }
}

/**
 * Realiza tratamento de entrada para o estado compartilhado do simulador
 * @param sm Gerenciador de estados
 * @param event Evento de entrada
 */
void  PlayingState::sharedInput(StateManager* sm, ALLEGRO_EVENT event) {
    if(event.type == ALLEGRO_EVENT_DISPLAY_CLOSE) {
        sm->quit();
    }
    if(event.type == ALLEGRO_EVENT_MOUSE_BUTTON_DOWN) {
        //printf("Pressed: %d %d\n", event.mouse.x,event.mouse.y);
        // Se apertou em algum local acima da GUI
        if(event.mouse.y <= gui->getPy()) {
            mouse = event.mouse.button;
            if(mouse == 1) {
                selectSX = mousex;
                selectSY = mousey;
                selectFX = mousex;
                selectFY = mousey;
            }
        } else {
            printf("Apertou: %s\n", gui->getAction().c_str());
            executeCommand(gui->getAction());
        }
    }
    if(event.type == ALLEGRO_EVENT_MOUSE_AXES) {
        mousex = event.mouse.x/zoom;
        mousey = event.mouse.y/zoom;
        mousex = mousex + scroll_x;
        mousey = mousey + scroll_y;
        if(mouse == 1) {
            selectFX = mousex;
            selectFY = mousey;
            //printf("Mouse(%d): %f %f\n", mouse, x,y);
        }

        if(mouse == 3) {
            float x = event.mouse.dx / zoom;
            float y = event.mouse.dy / zoom;
            scroll_x -= x * cos(rotate) + y * sin(rotate);
            scroll_y -= y * cos(rotate) - x * sin(rotate);
            //printf("Mouse(%d): %f %f\n", mouse, x,y);
        }
        zoom += event.mouse.dz * 0.1 * zoom;
        if (zoom < 0.1) zoom = 0.1;
        if (zoom > 10) zoom = 10;
    }
    if(event.type == ALLEGRO_EVENT_KEY_DOWN) {
        if(event.keyboard.keycode == ALLEGRO_KEY_F1) {
            debug[KEY::F1] = true;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F2) {
            debug[KEY::F2] = true;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F3) {
            debug[KEY::F3] = true;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F4) {
            debug[KEY::F4] = true;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F5) {
            debug[KEY::F5] = true;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F6) {
            debug[KEY::F6] = true;
        }
    }
    if(event.type == ALLEGRO_EVENT_KEY_UP) {
        if(event.keyboard.keycode == ALLEGRO_KEY_F1) {
            debug[KEY::F1] = false;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F2) {
            debug[KEY::F2] = false;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F3) {
            debug[KEY::F3] = false;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F4) {
            debug[KEY::F4] = false;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F5) {
            debug[KEY::F5] = false;
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_F6) {
            debug[KEY::F6] = false;
        }

    }

    if(event.type == ALLEGRO_EVENT_TIMER) {
        redraw = true;
    }
    if(event.type == ALLEGRO_EVENT_DISPLAY_RESIZE) {
        al_acknowledge_resize(display_);
        redraw = true;
    }
}

/**
 * Realiza tratamento de entrada para o estado IDLE do simulador
 * @param sm Gerenciador de estados
 * @param event Evento de entrada
 */
void PlayingState::idleInput(StateManager* sm, ALLEGRO_EVENT event) {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    std::vector<Building*> buildingSet = game_->getBuildingSet();
    if(event.type == ALLEGRO_EVENT_KEY_DOWN) {
        if (event.keyboard.keycode == ALLEGRO_KEY_ESCAPE) {
            sm->changeState(PauseState::getInstance(display_,timer_,queue_));
        }
    }
    if(event.type == ALLEGRO_EVENT_MOUSE_BUTTON_UP) {
        if(mouse == 1) {
            // Re-selecao permitida somente no estado idle, sel_unit, sel_building
            // Seleciona unidades dentro do retangulo
            for(int i = 0; i < (int) unitSet.size(); i++) {
                Unit* u = unitSet[i];
                if(mymath::inside(u->getPx(), u->getPy(), u->getRadius(), selectSX, selectSY, selectFX, selectFY)) {
                    select.push_back(u);
                }
            }
            if(select.size() > 0) {
                // Ordena select em ordem de type
                std::sort(select.begin(), select.end(), unitTypeSort);
            }
            // Se nao selecionou nenhuma unidade, busca pelas construcoes
            if(select.size() == 0) {
                for(int i = 0; i < (int) buildingSet.size(); i++) {
                    Building* b = buildingSet[i];
                    if(mymath::inside(b->getPx(), b->getPy(), b->getRadius(), selectSX, selectSY, selectFX, selectFY)) {
                        selectBuilding = b;
                        break;
                    }
                }
            }
            // Envia selecionados para gui apenas para mostrar suas informacoes na tela
            gui->setSelect(select);
            gui->setSelectBuilding(selectBuilding);
            // Muda de estado de acordo com o que foi selecionado
            if(select.size() > 0) {
                changeState(ACTION_STATE::SELECT_UNIT);
            } else if(selectBuilding != NULL) {
                changeState(ACTION_STATE::SELECT_BUILDING);
            }
        }
        mouse = 0;
        selectSX = selectSY = selectFX = selectFY = 0;
    }
}

/**
 * Realiza tratamento de entrada para o estado SELECT_UNIT do simulador
 * @param sm Gerenciador de estados
 * @param event Evento de entrada
 */
void PlayingState::selectUnitInput(StateManager* sm, ALLEGRO_EVENT event) {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    std::vector<Building*> buildingSet = game_->getBuildingSet();
    if(event.type == ALLEGRO_EVENT_KEY_DOWN) {
        if(event.keyboard.keycode == ALLEGRO_KEY_ESCAPE) {
            //action_state.pop();
        }
        if(event.keyboard.keycode == ALLEGRO_KEY_A) {
            executeCommand("Attack");
        }
    }
    if(event.type == ALLEGRO_EVENT_MOUSE_BUTTON_DOWN) {
        // Se apertou em algum local acima da GUI
        if(event.mouse.y <= gui->getPy()) {
            if(mouse == 2) {
                game_->multiAction(select, mousex, mousey);
                //game_->selectMove(select, mousex, mousey);
            }
        }
    }

    if(event.type == ALLEGRO_EVENT_MOUSE_BUTTON_UP) {
        if(mouse == 1) {
            select.clear();
            // Re-selecao permitida somente no estado idle, sel_unit, sel_building
            // Seleciona unidades dentro do retangulo
            for(int i = 0; i < (int) unitSet.size(); i++) {
                Unit* u = unitSet[i];
                if(mymath::inside(u->getPx(), u->getPy(), u->getRadius(), selectSX, selectSY, selectFX, selectFY)) {
                    select.push_back(u);
                }
            }
            if(select.size() > 0) {
                // Ordena select em ordem de type
                std::sort(select.begin(), select.end(), unitTypeSort);
            }
            // Se nao selecionou nenhuma unidade, busca pelas construcoes
            if(select.size() == 0) {
                for(int i = 0; i < (int) buildingSet.size(); i++) {
                    Building* b = buildingSet[i];
                    if(mymath::inside(b->getPx(), b->getPy(), b->getRadius(), selectSX, selectSY, selectFX, selectFY)) {
                        selectBuilding = b;
                        break;
                    }
                }
            }
            // Envia selecionados para gui apenas para mostrar suas informacoes na tela
            gui->setSelect(select);
            gui->setSelectBuilding(selectBuilding);
            // Muda de estado de acordo com o que foi selecionado
            if(select.size() > 0) {
                action_state.pop();
                changeState(ACTION_STATE::SELECT_UNIT);
            } else if(selectBuilding != NULL) {
                action_state.pop();
                changeState(ACTION_STATE::SELECT_BUILDING);
            } else {
                //changeState(ACTION_STATE::IDLE);
                action_state.pop();
            }
        }
        mouse = 0;
        selectSX = selectSY = selectFX = selectFY = 0;
    }
}

/**
 * Realiza tratamento de entrada para o estado SELECT_BUILDING do simulador
 * @param sm Gerenciador de estados
 * @param event Evento de entrada
 */
void PlayingState::selectBuildingInput(StateManager* sm, ALLEGRO_EVENT event) {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    std::vector<Building*> buildingSet = game_->getBuildingSet();
    if(event.type == ALLEGRO_EVENT_KEY_DOWN) {
        if(event.keyboard.keycode == ALLEGRO_KEY_ESCAPE) {
            game_->cancelBuildingQueue(selectBuilding);
        }
    }
    if (event.type == ALLEGRO_EVENT_MOUSE_BUTTON_DOWN) {
        // Se apertou em algum local acima da GUI
        if(event.mouse.y <= gui->getPy()) {
            if(mouse == 2) {
                game_->buildingPath(selectBuilding, mousex, mousey);
            }
        }
    }
    if(event.type == ALLEGRO_EVENT_MOUSE_BUTTON_UP) {
        if(mouse == 1) {
            selectBuilding = NULL;
            // Re-selecao permitida somente no estado idle, sel_unit, sel_building
            // Seleciona unidades dentro do retangulo
            for(int i = 0; i < (int) unitSet.size(); i++) {
                Unit* u = unitSet[i];
                if(mymath::inside(u->getPx(), u->getPy(), u->getRadius(), selectSX, selectSY, selectFX, selectFY)) {
                    select.push_back(u);
                }
            }
            if(select.size() > 0) {
                // Ordena select em ordem de type
                std::sort(select.begin(), select.end(), unitTypeSort);
            }
            // Se nao selecionou nenhuma unidade, busca pelas construcoes
            if(select.size() == 0) {
                for(int i = 0; i < (int) buildingSet.size(); i++) {
                    Building* b = buildingSet[i];
                    if(mymath::inside(b->getPx(), b->getPy(), b->getRadius(), selectSX, selectSY, selectFX, selectFY)) {
                        selectBuilding = b;
                        break;
                    }
                }
            }
            // Envia selecionados para gui apenas para mostrar suas informacoes na tela
            gui->setSelect(select);
            gui->setSelectBuilding(selectBuilding);
            // Muda de estado de acordo com o que foi selecionado
            if(select.size() > 0) {
                action_state.pop();
                changeState(ACTION_STATE::SELECT_UNIT);
            } else if(selectBuilding != NULL) {
                action_state.pop();
                changeState(ACTION_STATE::SELECT_BUILDING);
            } else {
                //changeState(ACTION_STATE::IDLE);
                action_state.pop();
            }
        }
        mouse = 0;
        selectSX = selectSY = selectFX = selectFY = 0;
    }
}

/**
 * Realiza tratamento de entrada para o estado TARGET do simulador
 * @param sm Gerenciador de estados
 * @param event Evento de entrada
 */
void PlayingState::targetInput(StateManager* sm, ALLEGRO_EVENT event) {
    static bool execute = false;
    std::vector<Unit*> unitSet = game_->getUnitSet();
    std::vector<Building*> buildingSet = game_->getBuildingSet();
    if(event.type == ALLEGRO_EVENT_KEY_DOWN) {
        if(event.keyboard.keycode == ALLEGRO_KEY_ESCAPE) {
            action_state.pop();
        }
    }
    if (event.type == ALLEGRO_EVENT_MOUSE_BUTTON_DOWN) {
        // Se apertou em algum local acima da GUI
        if(event.mouse.y <= gui->getPy()) {
            if(mouse == 1) {
                runCommand(command, mousex, mousey);
                execute = true;
            }
            if(mouse == 2) {
                game_->multiAction(select, mousex, mousey);
                action_state.pop();
            }
        }
    }
    if(event.type == ALLEGRO_EVENT_MOUSE_BUTTON_UP) {
        if(execute) {
            action_state.pop();
            execute = false;
        }
        mouse = 0;

    }
}

/**
 * Executa o tratamento de entrada dependendo do estado atual do simulador
 * @param sm Gerenciador de estados
 */
void PlayingState::readInput(StateManager* sm) {
    //Graph* G = game_->getGraph();
    //Node*** node = G->getNode();
    ALLEGRO_EVENT event;
    al_wait_for_event(queue_, &event);
    sharedInput(sm, event);
    if(action_state.top() == ACTION_STATE::IDLE) {
        idleInput(sm, event);
    } else if(action_state.top() == ACTION_STATE::SELECT_UNIT) {
        selectUnitInput(sm, event);
    } else if(action_state.top() == ACTION_STATE::SELECT_BUILDING) {
        selectBuildingInput(sm, event);
    } else if(action_state.top() == ACTION_STATE::TARGET) {
        targetInput(sm, event);
    }
    // Le comandos feitos pela IA
    for(int i = 0; i < NUM_AI; i++) {
        if(ai[i] != NULL) ai[i]->run();
    }
}

/**
 * Renderiza tela principal do jogo
 * @param sm Gerenciador de estados
 */
void PlayingState::render(StateManager* sm) {
    if(redraw && al_is_event_queue_empty(queue_)) {
        redraw = false;
        Map* map = game_->getMap();
        ALLEGRO_TRANSFORM transform;

        al_identity_transform(&transform);
        al_translate_transform(&transform, -scroll_x, -scroll_y);
        al_scale_transform(&transform, zoom, zoom);
        al_use_transform(&transform);
        al_clear_to_color(al_map_rgb(0, 0, 0));

        // Render dos tiles
        map->render();
        // Render das unidades, cor do time, barra de hp
        std::vector<Unit*> unitSet = game_->getUnitSet();
        for(int i = 0; i < (int) unitSet.size(); i++) {
            // Cor do time
            al_draw_filled_circle(unitSet[i]->getPx(), unitSet[i]->getPy(), unitSet[i]->getRadius(), game_->getPlayer()[unitSet[i]->getTeam()]->getColor());
            unitSet[i]->render();
        }
        // Render das construcoes, cor do time
        std::vector<Building*> buildingSet = game_->getBuildingSet();
        for(int i = 0; i < (int) buildingSet.size(); i++) {
            // Cor do time
            al_draw_filled_circle(buildingSet[i]->getPx(), buildingSet[i]->getPy(), buildingSet[i]->getRadius(), game_->getPlayer()[buildingSet[i]->getTeam()]->getColor());
            buildingSet[i]->render();
        }
        std::vector<Projectile*> projectileSet = game_->getProjectileSet();
        for(int i = 0; i < (int) projectileSet.size(); i++) {
            projectileSet[i]->render();
        }
        // Render da selecao de unidade
        if (select.size() > 0) {
            for(int i = 0; i < (int) select.size(); i++) {
                Unit* u = select[i];
                al_draw_circle(u->getPx(), u->getPy(), u->getRadius(), al_map_rgba(0,255,0, 128), 1);
                //al_draw_arc(u->getPx(), u->getPy(), u->getRadius(),-M_PI/4,M_PI+M_PI/2, al_map_rgba(0,255,0, 128), 1.5);
                // Render do pixelpath
                if(select[i]->getPixelPath().size() > 0) {
                    int raio = 2;
                    for(int j = 0; j < (int) select[i]->getPixelPath().size(); j++) {
                        al_draw_filled_circle(select[i]->getPixelPath()[j].first,select[i]->getPixelPath()[j].second,raio,al_map_rgb(10,25,0));
                    }
                }
                // Render do collisionpath
                if(select[i]->getCollisionPath().size() > 0) {
                    int raio = 1;
                    for(int j = 0; j < (int) select[i]->getCollisionPath().size(); j++) {
                        al_draw_filled_circle(select[i]->getCollisionPath()[j].first,select[i]->getCollisionPath()[j].second,raio,al_map_rgb(255,0,0));
                    }
                }

            }
        }
        // Render na selecao de construcao
        if(selectBuilding != NULL) {
            Building* b = selectBuilding;
            al_draw_circle(b->getPx(), b->getPy(), b->getRadius(), al_map_rgba(0,255,0,128), 1);
            //al_draw_arc(b->getPx(), b->getPy(), b->getRadius(),-M_PI/4,M_PI+M_PI/2, al_map_rgba(0,255,0, 128), 1.5);
            if(b->getPixelPath().size() > 0) {
                int raio = 2;
                for(int j = 0; j < (int) b->getPixelPath().size(); j++) {
                    al_draw_filled_circle(b->getPixelPath()[j].first,b->getPixelPath()[j].second,raio,al_map_rgb(10,25,0));
                }
            }
        }
        // Render no fantasma do mouseover da construcao
        if(action_state.top() == ACTION_STATE::TARGET && buildingID != -1) {
            BuildingProperty* bp = BuildingProperty::getInstance();
            int x = mymath::div(mousex,TILE_WIDTH);
            int y = mymath::div(mousey,TILE_HEIGHT);
            x = x - bp->getDimension(buildingID)/2;
            y = y - bp->getDimension(buildingID)/2;
            al_draw_bitmap(bp->getBitmap(buildingID), x*TILE_WIDTH, y*TILE_HEIGHT, 0);
        }
        // Render do retangulo da selecao
        if(mouse == 1) {
            al_draw_filled_rectangle(selectSX,selectSY,selectFX, selectFY,al_map_rgba(0,64,0,0));
        }
        // Render Quadtree
        if(debug[KEY::F1]) {
            std::queue<Quadtree*> queue;
            queue.push(collision->getQt());
            Quadtree* atual;
            while(!queue.empty()) {
                atual = queue.front();
                queue.pop();
                for(int i = 0; i < 4; i++) {
                    if(atual->getNodes()[i] != NULL) {
                        queue.push(atual->getNodes()[i]);
                    }
                }
                Rect* ret = atual->getBounds();
                al_draw_rectangle(ret->getX(), ret->getY(), ret->getX()+ret->getWidth(), ret->getY()+ ret->getHeight(),al_map_rgb(255,255,255),2);
            }
        }
        // Render zona de equivalencia
        if(debug[KEY::F2]) {
            for(int i = 0; i < game_->getMap()->getRows(); i++) {
                for(int j = 0; j < game_->getMap()->getColumns(); j++) {
                    Node*** node = game_->getGraph()->getNode();
                    al_draw_textf(font_gui, al_map_rgb(0,0,0), i*TILE_WIDTH, j*TILE_HEIGHT, ALLEGRO_ALIGN_LEFT, "%d", game_->getGraph()->getZoneEquivalency()[node[i][j]->getZone()]);
                }
            }
        }
        // Render no buildType
        if(debug[KEY::F3]) {
            for(int i = 0; i < game_->getMap()->getRows(); i++) {
                for(int j = 0; j < game_->getMap()->getColumns(); j++) {
                    Node*** node = game_->getGraph()->getNode();
                    al_draw_textf(font_gui, al_map_rgb(0,0,0), i*TILE_WIDTH, j*TILE_HEIGHT, ALLEGRO_ALIGN_LEFT, "%d", node[i][j]->getTileType());
                }
            }
        }
        // Render dos nós visitados no pathfinding
        if(debug[KEY::F4]) {
            for(int i = 0; i < game_->getMap()->getRows(); i++) {
                for(int j = 0; j < game_->getMap()->getColumns(); j++) {
                    Node*** node = game_->getGraph()->getNode();
                    if(node[i][j]->getParent() != NULL) {
                        //al_draw_filled_rectangle(i*TILE_WIDTH,j*TILE_HEIGHT,i*TILE_WIDTH+TILE_WIDTH, j*TILE_HEIGHT+TILE_HEIGHT,al_map_rgba(0,128,0,0));
                        ALLEGRO_COLOR color = al_map_rgb(255,255,255);
                        if(node[i][j]->isOnPath()) color = al_map_rgb(255,0,0);
                        renderArrow(i*TILE_WIDTH+TILE_WIDTH/2, j*TILE_HEIGHT+TILE_HEIGHT/2, node[i][j]->getParent()->getX()-i, node[i][j]->getParent()->getY()-j, color);
                    }
                }
            }
        }
        // Render nos pontos de estrangulamento
        if(debug[KEY::F5]) {
            for(int i = 0; i < game_->getMap()->getRows(); i++) {
                for(int j = 0; j < game_->getMap()->getColumns(); j++) {
                    Node*** node = game_->getGraph()->getNode();
                    int choke = node[i][j]->getChoke();
                    if(choke >= 0) {
                        //al_draw_textf(font_gui, al_map_rgb(0,0,0), i*TILE_WIDTH, j*TILE_HEIGHT, ALLEGRO_ALIGN_LEFT, "%c", 'C');
                        //al_draw_filled_rectangle(i*TILE_WIDTH,j*TILE_HEIGHT,i*TILE_WIDTH+TILE_WIDTH,j*TILE_HEIGHT+TILE_HEIGHT,al_map_rgba(255,0,0,128));
                        al_draw_filled_circle(i*TILE_WIDTH+TILE_WIDTH/2,j*TILE_HEIGHT+TILE_HEIGHT/2,5, game_->getPlayer()[choke]->getColor());
                    }
                }
            }
        }

        // Render no ponto de ataque da IA
        if(debug[KEY::F6]) {
            for(int i = 0; i < NUM_AI; i++) {
                al_draw_filled_circle(ai[i]->infoManager_->attackPoint.first, ai[i]->infoManager_->attackPoint.second,5, game_->getPlayer()[i]->getColor());

            }
            /*
            for(int i = 0; i < game_->getMap()->getRows(); i++) {
                for(int j = 0; j < game_->getMap()->getColumns(); j++) {
                    Node*** node = game_->getGraph()->getNode();
                    al_draw_textf(font_gui, al_map_rgb(0,0,0), i*TILE_WIDTH, j*TILE_HEIGHT, ALLEGRO_ALIGN_LEFT, "%d", node[i][j]->getTeam());
                }
            }
            */
        }

        al_hold_bitmap_drawing(0);
        al_identity_transform(&transform);
        al_use_transform(&transform);
        if (font) {
            //al_draw_textf(font, al_map_rgb(255, 255, 255), 0, 0, ALLEGRO_ALIGN_LEFT, "FPS: %d", fps);
        }
        gui->render();
        al_flip_display();
        double t = al_get_time();
        //printf("getTIME: %f\n", t);
        fps_accum++;
        if (t - fps_time >= 1) {
            fps = fps_accum;
            fps_accum = 0;
            fps_time = t;
            //UnitProperty::getInstance()->fps = fps; // RETIRE ESSE COMENTARIO PARA ALTERAR A VELOCIDADE DAS UNIDADES DE ACORDO COM O FPS ATUAL
        }
        //printf("fps: %d, fps_acumm: %d, fps_time: %f\n", fps, fps_accum, fps_time);
    }
}

/**
 * Atualiza simulador
 * @param sm Gerenciador de estados
 */
void PlayingState::update(StateManager* sm) {
    Player* winner = NULL;
    if(redraw == true && al_is_event_queue_empty(queue_)) {
        // Verifica condicao de vitoria
        if((winner = getWinner()) != NULL) {
            ResultState::getInstance(display_,timer_,queue_)->setWinner(winner);
            sm->changeState(ResultState::getInstance(display_,timer_,queue_));
        }
        // Atualiza movimentacao das unidades
        updateUnit();
        // Trata colisao
        updateCollision();
        // Atualiza ataques das unidades
        updateAttack();
        // Atualiza projeteis
        updateProjectile();
        // Atualiza ponteiros
        updatePointer();
        // Remove mortos do jogo
        updateHp();
        // Para cada construcao verifica se o tempo de treinamento ja chegou no tempo necessario de producao da unidade
        updateBuilding();
        // Atualiza worker
        updateWorker();
        // Atualiza construcao
        updateIteration();
        // Atualiza GUI
        gui->update(action_state.top());

    }
}

/**
 * Atualiza movimentacao das unidades
 */
void PlayingState::updateUnit() {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    /* Atualiza unidades */
    for(int i = 0; i < (int) unitSet.size(); i++) {
        if(unitSet[i]->isRepath()) {
            unitSet[i]->setRepath(false);
            float temptx = -1;
            float tempty = -1;
            if(unitSet[i]->getFollow() != NULL) {
                temptx = unitSet[i]->getFollow()->getPx();
                tempty = unitSet[i]->getFollow()->getPy();
            } else {
                temptx = unitSet[i]->getTx();
                tempty = unitSet[i]->getTy();
            }
            unitSet[i]->resetCollision();
            game_->unitMove(unitSet[i], temptx, tempty);
        }
        if(unitSet[i]->getFollow() != NULL) {
            int xf = mymath::div(unitSet[i]->getFollow()->getPx(),TILE_WIDTH);
            int yf = mymath::div(unitSet[i]->getFollow()->getPy(),TILE_HEIGHT);
            if(xf != unitSet[i]->getXFollow() || yf != unitSet[i]->getYFollow()) {
                //printf("f(%d %d) , current(%d %d)\n", xf, yf, unitSet[i]->getXFollow(),unitSet[i]->getYFollow());
                unitSet[i]->setXFollow(xf);
                unitSet[i]->setYFollow(yf);
                game_->unitMove(unitSet[i], unitSet[i]->getFollow()->getPx(), unitSet[i]->getFollow()->getPy());
            }
        }
        unitSet[i]->update();
    }
}

/**
 * Atualiza colisao
 */
void PlayingState::updateCollision() {
    collision->solve(game_->getUnitSet());
}

//Faz atualizacoes de ataque em unidades e construcoes
//A prioridade de ataque se estiver no alcance do atacante
//1) Alvo focus
//2) Alvo unidade do frame anterior
//3) Alvo unidade mais proximo
//4) Alvo construcao mais proximo
/**
 * Atualiza ataques das unidades
 */
void PlayingState::updateAttack() {
    Unit* unitA;
    //Entity* target = NULL;
    Entity* unitTarget;
    Entity* buildingTarget;
    Entity* lastTarget;
    Entity* lastFollow;
    Entity* focus;
    std::vector<Unit*> unitSet = game_->getUnitSet();
    for(int i = 0; i < (int) unitSet.size(); i++) {
        unitA = unitSet[i];
        focus = unitA->getFocus();
        lastTarget = unitA->getTarget();
        lastFollow = unitA->getFollow();
        bool following = false;
        unitA->setHold(false);
        if(!game_->getGraph()->getNode()[(int) (unitA->getPx()/TILE_WIDTH)][(int) (unitA->getPy()/TILE_HEIGHT)]->isWalkable()) continue;
        if(unitA->isAgressive()) {
            unitA->setTarget(NULL);
            if(focus != NULL && focus->getHp() > 0 && mymath::inrange(unitA->getPx(), unitA->getPy(), unitA->getRange(), focus->getPx(), focus->getPy(), focus->getRadius())) {
                unitA->setTarget(focus);
            } else if(focus == NULL) {
                if(lastTarget != NULL && lastTarget->isUnit() && mymath::inrange(unitA->getPx(), unitA->getPy(), unitA->getRange(), lastTarget->getPx(), lastTarget->getPy(), lastTarget->getRadius())) {
                    unitA->setTarget(lastTarget);
                } else if((unitTarget = game_->getEnemyUnit(unitA)) != NULL &&
                          mymath::inrange(unitA->getPx(), unitA->getPy(), unitA->getRange(), unitTarget->getPx(), unitTarget->getPy(), unitTarget->getRadius())) {
                    unitA->setTarget(unitTarget);
                } else if((buildingTarget = game_->getEnemyBuilding(unitA)) != NULL &&
                          mymath::inrange(unitA->getPx(), unitA->getPy(), unitA->getRange(), buildingTarget->getPx(), buildingTarget->getPy(), buildingTarget->getRadius())) {
                    unitA->setTarget(buildingTarget);
                } else if(lastFollow != NULL && lastFollow->isUnit() && mymath::inrange(unitA->getPx(), unitA->getPy(), unitA->getAggroRadius(), lastFollow->getPx(), lastFollow->getPy(), lastFollow->getRadius())) {
                    following = true;
                } else if(unitTarget != NULL && mymath::inrange(unitA->getPx(), unitA->getPy(), unitA->getAggroRadius(), unitTarget->getPx(), unitTarget->getPy(), unitTarget->getRadius())) {
                    //game_->unitMove(unitA, unitTarget->getPx(), unitTarget->getPy());
                    game_->aFollowb(unitA, unitTarget);
                    //unitA->setTarget(NULL);
                    following = true;
                } else if(buildingTarget != NULL && mymath::inrange(unitA->getPx(), unitA->getPy(), unitA->getAggroRadius(), buildingTarget->getPx(), buildingTarget->getPy(), buildingTarget->getRadius())) {
                    //game_->unitMove(unitA, buildingTarget->getPx(), buildingTarget->getPy());
                    game_->aFollowb(unitA, buildingTarget);
                    //unitA->setTarget(NULL);
                    following = true;
                }
            }
            if(!following && unitA->getFollow() != NULL && unitA->getFollow() != unitA->getFocus()) {
                game_->stopFollow(unitA);
            }
            if(unitA->getTarget() != NULL) {
                unitA->wait();
                unitA->setHold(true);
                if(unitA->isAttackReady()) {
                    //unitA->getTarget()->receiveDamage(unitA->getDamage());
                    if(unitA->getProjectile() == PROJECTILE_TYPE::MELEE) {
                        unitA->getTarget()->receiveDamage(unitA->getDamage());
                    } else {
                        Projectile* p = new Projectile(unitA->getPx(), unitA->getPy(), unitA->getDamage(), unitA->getProjectile(), unitA->getTarget());
                        game_->insertProjectile(p);
                    }
                    unitA->setCooldown(unitA->getCooldown()-unitA->getAttackSpeed());
                    unitA->actionSprite();
                }
            }
        } else {
            unitA->setTarget(NULL);
        }
    }
}

// Atualiza projeteis, se o projetil chegou no target, faz o target receber dano
/**
 * Atualiza movimentacao e acao dos projeteis
 */
void PlayingState::updateProjectile() {
    std::vector<Projectile*> proj = game_->getProjectileSet();
    for(int i = 0; i < (int) proj.size(); i++) {
        proj[i]->update();
        Entity* target = proj[i]->getTarget();
        if(target != NULL) {
            if(mymath::inrange(proj[i]->getPx(), proj[i]->getPy(), proj[i]->getRadius(), target->getPx(), target->getPy(), target->getRadius())) {
                target->receiveDamage(proj[i]->getDamage());
                game_->removeProjectile(proj[i]);
            }
        }
    }
}

// Para cada unidade, se target/focus estiverem morto entao muda o target pra NULL
//   Se andou todo caminho ate o ponto inicial de focus, seta focus pra NULL
/**
 * Atualiza ponteiros das entidades
 */
void PlayingState::updatePointer() {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    std::vector<Building*> buildingSet = game_->getBuildingSet();
    for(int i = 0; i < (int) unitSet.size(); i++) {
        Unit* u = unitSet[i];
        if(u->getTarget() != NULL && u->getTarget()->getHp() <= 0) {
            u->setTarget(NULL);
        }
        if(u->getFocus() != NULL && u->getFocus()->getHp() <= 0) {
            u->setFocus(NULL);
        }
        if(u->getEntity() != NULL && u->getEntity()->getHp() <= 0) {
            u->setEntity(NULL);
        }
        if(u->getFollow() != NULL && u->getFollow()->getHp() <= 0) {
            u->setFollow(NULL);
        }
    }
    for(int i = 0; i < (int) buildingSet.size(); i++) {
        Building* b = buildingSet[i];
        if(b->getEntity() != NULL && b->getEntity()->getHp() <= 0) {
            b->setEntity(NULL);
        }
    }
    for(int i = 0; i < (int) select.size(); i++) {
        if(select[i]->getHp() <= 0) {
            select.erase(select.begin() + i);
            i--;
        }
        if(select.size() == 0) {
            action_state.pop();
        }
    }
    if(selectBuilding != NULL && selectBuilding->getHp() <= 0) {
        selectBuilding = NULL;
        action_state.pop();
    }
    gui->setSelect(select);
    gui->setSelectBuilding(selectBuilding);
}

/**
 * Atualiza vida das entidades e remove se for o caso
 */
void PlayingState::updateHp() {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    std::vector<Building*> buildingSet = game_->getBuildingSet();
    std::vector<Projectile*> proj = game_->getProjectileSet();
    for(int i = 0; i < (int) proj.size(); i++) {
        if(proj[i]->getTarget()->getHp() <= 0) {
            game_->removeProjectile(proj[i]);
        }
    }
    // Os dois laços abaixo enviam sinal para AI que a unidade/construcao foram removidos do jogo
    for(int i = 0; i < (int) unitSet.size(); i++) {
        if(unitSet[i]->getHp() <= 0) {
            // Para cada IA, remove Unidade
            for(int j = 0; j < NUM_AI; j++) {
                if(ai[j] != NULL/* && ai[j]->getTeam() == unitSet[i]->getTeam()*/) {
                    ai[j]->removeUnit(unitSet[i]);
                }
            }
            game_->removeUnit(unitSet[i]);
        }
    }
    for(int i = 0; i < (int) buildingSet.size(); i++) {
        if(buildingSet[i]->getHp() <= 0) {
            // Para cada IA, remove Construcao
            for(int j = 0; j < NUM_AI; j++) {
                if(ai[j] != NULL/* && ai[j]->getTeam() == buildingSet[i]->getTeam()*/) {
                    // Se a construcao tiver unidade dentro, remove da IA
                    for(int k = 0; k < (int) buildingSet[i]->getSlot().size(); k++) {
                        ai[j]->removeUnit(buildingSet[i]->getSlot()[k]);
                    }
                    ai[j]->removeBuilding(buildingSet[i]);
                }
            }
            game_->removeBuilding(buildingSet[i]);
        }
    }
}

/**
 * Atualiza tempo de treinamento e coleta de recurso de cada edificio
 */
void PlayingState::updateBuilding() {
    Unit *u;
    Building* b;
    std::vector<std::pair<float,float>> pixelPath;
    std::pair<float,float> lastPath;
    std::vector<Building*> buildingSet = game_->getBuildingSet();
    for(int i = 0; i < (int) buildingSet.size(); i++) {
        b = buildingSet[i];
        b->update();
        if(b->getQueue().size() > 0) {
            float trainingTime = b->getTrainingTime();
            float unitTime = UnitProperty::getInstance()->getTime(b->getQueue()[0]);
            if(trainingTime >= unitTime) {
                u = new Unit(b->getPx(),b->getPy(), b->getQueue()[0], b->getTeam());
                pixelPath = b->getPixelPath();
                lastPath = pixelPath[pixelPath.size()-1];
                u->setT(lastPath.first, lastPath.second);
                u->setPixelPath(b->getPixelPath());
                u->setEntity(b->getEntity());
                u->setAgressive(true);
                game_->insertUnit(u);
                // Insere na AI
                for(int j = 0; j < NUM_AI; j++) {
                    if(ai[j] != NULL/* && ai[j]->getTeam() == u->getTeam()*/) {
                        ai[j]->insertUnit(u);
                    }
                }
                b->removeQueue(0);
                b->setTrainingTime(trainingTime-unitTime);
            }
        }
        if(b->getSlot().size() > 0) {
            float miningTime = b->getMiningTime();
            float increaseTime = b->getMiningCooldown();
            int increaseGold = b->getMiningGain();
            //printf("%f\n", miningTime);
            if(miningTime >= increaseTime) {
                std::vector<Player*> player = game_->getPlayer();
                player[b->getTeam()]->sumGold(increaseGold);
                //printf("incrementou\n");
                b->setMiningTime(miningTime-increaseTime);
            }
        }
    }
}

/**
 * Atualiza unidades trabalhadoras, faz o tratamento no caso da unidade trabalhadora ter chegado ao ponto de onde ele quer construir
 */
void PlayingState::updateWorker() {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    for(int i = 0; i < (int) unitSet.size(); i++) {
        Unit* u = unitSet[i];
        if(u->getBuild() != -1 && u->getType() == UNIT_TYPE::WORKER) {
            /*
            std::vector<std::pair<float,float>> pixelPath = u->getPixelPath();
            float tx = pixelPath[pixelPath.size()-1].first;
            float ty = pixelPath[pixelPath.size()-1].second;
            */
            float tx = u->getTx();
            float ty = u->getTy();
            int x = mymath::div(tx,TILE_WIDTH);
            int y = mymath::div(ty,TILE_HEIGHT);
            float px = x*TILE_WIDTH + TILE_WIDTH/2;
            float py = y*TILE_HEIGHT + TILE_HEIGHT/2;
            x = x - BuildingProperty::getInstance()->getDimension(u->getBuild())/2;
            y = y - BuildingProperty::getInstance()->getDimension(u->getBuild())/2;
            if(mymath::distance(u->getPx(),u->getPy(),px,py) <= BuildingProperty::getInstance()->getRadius(u->getBuild()) + UnitProperty::getInstance()->getRange(u->getType())) {
                u->wait();
                if(!game_->isInsertBuildingValid(tx,ty,u->getBuild())) {
                    game_->removeUnitBuild(u);
                } else if(game_->hasUnitInBuilding(x,y,u->getBuild())) {
                    game_->removeUnitBuild(u);
                } else {
                    Building* b = new Building(x, y, u->getBuild(), u->getTeam());
                    game_->insertBuilding(b);
                    u->setEntity(b);
                    u->setBuild(-1);
                    // Insere na AI
                    for(int j = 0; j < NUM_AI; j++) {
                        if(ai[j] != NULL/* && ai[j]->getTeam() == b->getTeam()*/) {
                            ai[j]->insertBuilding(b);
                        }
                    }
                }
                u->stop();
                u->setAgressive(false);
            }
        }
    }
}

/**
 * Atualiza iteracao da unidade trabalhadora com os edificios
 */
void PlayingState::updateIteration() {
    std::vector<Unit*> unitSet = game_->getUnitSet();
    for(int i = 0; i < (int) unitSet.size(); i++) {
        Unit* u = unitSet[i];
        Building* b;
        if(u->getEntity() == NULL) continue;
        if(!u->getEntity()->isBuilding()) continue;
        b = (Building*) u->getEntity();
        if(u->getType() == UNIT_TYPE::WORKER) {
            if(mymath::distance(u->getPx(),u->getPy(),b->getPx(),b->getPy()) <= b->getRadius() + u->getRange()) {
                if(!b->isActive()) {
                    if(!b->isConstructing()) {
                        b->setConstructing(true);
                        u->actionSprite();
                        u->setHold(true);
                    }
                }
                else if(b->getType() == BuildingProperty::getInstance()->getType("Gold_Mine")) {
                    if(game_->insertGoldSlot(b, u)) {
                        std::vector<Unit*>::iterator it;
                        it = find(select.begin(),select.end(), u);
                        if(it != select.end()) {
                            select.erase(it);
                        }
                    }
                }
                if(b->isActive()) {
                    u->setEntity(NULL);
                    u->setAgressive(true);
                }
            }
        }
    }
}

/**
 * Interpreta uma string e roda um comando com
 * @param com String do comando
 */
void PlayingState::executeCommand(std::string com) {
    //Unit *u;
    int id;
    std::vector<std::pair<float,float>> pixelPath;
    std::pair<float,float> lastPath;
    command = com;
    buildingID = -1;
    if(com.compare("Move") == 0) {
        changeState(ACTION_STATE::TARGET);
    } else if(com.compare("Attack") == 0) {
        changeState(ACTION_STATE::TARGET);
    } else if(com.compare("Gather") == 0) {
        changeState(ACTION_STATE::TARGET);
    } else if(com.compare("Rally") == 0) {
        changeState(ACTION_STATE::TARGET);
    } else if((id = BuildingProperty::getInstance()->getType(com)) != -1) {
        buildingID = id;
        changeState(ACTION_STATE::TARGET);
    }
    if(action_state.top() != ACTION_STATE::TARGET) {
        runCommand(command);
    }
}

/**
 * Roda um comando com as entidades selecionadas
 * @param com String do comando
 */
void PlayingState::runCommand(std::string com) {
    int id;
    if(com.compare("Stop") == 0) {
        //game_->selectStop(select);
        game_->executeCommand(com, select);
    } else if(com.compare("Unload") == 0) {
        //game_->unloadGoldSlot(selectBuilding);
        game_->executeCommand(com, selectBuilding);
    } else if((id = UnitProperty::getInstance()->getType(com)) != -1) {
        //game_->insertBuildingQueue(selectBuilding, id);
        game_->executeCommand(com, selectBuilding);
    }

}

/**
 * Roda um comando no ponto (px,py) com as entidades selecionadas
 * @param com String do comando
 * @param px Coordenada x em pixel
 * @param py Coordenada y em pixel
 */
void PlayingState::runCommand(std::string com, float px, float py) {
    int id;
    if(com.compare("Move") == 0) {
        //game_->selectMove(select,px,py);
        game_->executeCommand(com, select, px, py);
    } else if(com.compare("Attack") == 0) {
        //game_->selectMoveAttack(select,px,py);
        game_->executeCommand(com, select, px, py);
    } else if(com.compare("Gather") == 0) {
        //game_->selectMoveGather(select,px,py);
        game_->executeCommand(com, select, px, py);
    } else if(com.compare("Rally") == 0) {
        //game_->buildingPath(selectBuilding, px, py);
        game_->executeCommand(com, selectBuilding, px, py);
    } else if((id = BuildingProperty::getInstance()->getType(com)) != -1) {
        //game_->selectMoveBuild(select, px, py, id);
        game_->executeCommand(com, select, px, py);
    }
}

/**
 * Muda estado do simulador
 * @param newState
 */
void PlayingState::changeState(int newState) {
    action_state.push(newState);
}

/**
 * Renderiza uma seta(vetor) apontando pra (x,y) na posicao (px,py)
 * @param px Coordenada x em pixel
 * @param py Coordenada y em pixel
 * @param x Valor do vetor na coordenada x
 * @param y Valor do vetor na coordenada y
 * @param color Cor da seta
 */
void PlayingState::renderArrow(int px, int py, int x, int y, ALLEGRO_COLOR color) {
    float fpx = px+x*TILE_WIDTH/3;
    float fpy = py+y*TILE_HEIGHT/3;
    al_draw_line(px-x*TILE_WIDTH/3, py-y*TILE_HEIGHT/3, fpx, fpy, color, 1);
    //al_draw_filled_circle(fpx,fpy, 2.5, al_map_rgb(255,255,255));
    if(y == 0) al_draw_filled_triangle(fpx, fpy, px, py+3, px, py-3, color);
    else if(x == 0) al_draw_filled_triangle(fpx, fpy, px+3, py, px-3, py, color);
    else al_draw_filled_triangle(fpx, fpy, px+x*2, py-y*2, px-x*2, py+y*2, color);
}

/**
 * Verifica se algum player ganhou o jogo(destruiu todos edificios inimigos)
 * @return Ponteiro do player vencedor, NULL se ninguem ganhou
 */
Player* PlayingState::getWinner() {
    int i;
    for(i = 0; i < (int) game_->getPlayer().size(); i++) {
        if(game_->getPlayer()[i]->getBuildingSet().size() == 0) {
            return game_->getPlayer()[1-i];
        }
    }
    return NULL;
}
