#include "Graph.h"

class mycomparison {
public:
    mycomparison() {
    }
    bool operator() (Node* lhs, Node* rhs) const {
        return (lhs->getCost() > rhs->getCost());
    }
};

/**
 * Construtor
 * @param mapMatrix Matriz do mapa
 */
Graph::Graph(std::vector<std::string> mapMatrix) {
    rows_ = mapMatrix.size();
    columns_ = mapMatrix[0].size();
    adj_ = new Edge**[rows_];
    node_ = new Node**[rows_];
    for(int i = 0; i < rows_; i++) {
        node_[i] = new Node*[columns_];
        adj_[i] = new Edge*[columns_];
    }
    for(int i = 0; i < rows_; i++) {
        for(int j = 0; j < columns_; j++) {
            node_[i][j] = new Node(i,j);
            adj_[i][j] = NULL;
        }
    }
    makeGraph(mapMatrix);
}

/**
 * Destrutor
 */
Graph::~Graph() {
    printf("Deletando node e adj de Graph\n");
    int i, j;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            delete node_[i][j];
            delete adj_[i][j];
        }
    }
    for(i = 0; i < rows_; i++) {
        delete [] node_[i];
        delete [] adj_[i];
    }
    delete [] adj_;
    delete [] node_;

}

/**
 * @return Quantidade de linhas da matriz do grafo
 */
int Graph::getRows() {
    return rows_;
}

/**
 * @return Quantidade de colunas da matriz do grafo
 */
int Graph::getColumns() {
    return columns_;
}

/**
 * @return Matriz de no's do grafo
 */
Node*** Graph::getNode() {
    return node_;
}

/**
 * @return Matriz contendo as arestas de cada no' do grafo
 */
Edge*** Graph::getAdj() {
    return adj_;
}

/**
 * @return Map das zonas de equivalencia
 */
std::map<int,int> Graph::getZoneEquivalency() {
    return zoneEquivalency_;
}

/**
 * Gera o grafo de acordo com o mapMatrix lido do arquivo
 * @param mapMatrix Matriz do mapa
 */
void Graph::makeGraph(std::vector<std::string> mapMatrix) {
    for(int i = 0; i < rows_; i++) {
        for(int j = 0; j < columns_; j++) {
            int type = TileProperty::getInstance()->getType(mapMatrix[i][j]);
            if(type == -1) {
                printf("Erro no makeGraph\n");
                return;
            }
            node_[i][j]->setType(type);
            node_[i][j]->setWalkable(TileProperty::getInstance()->isWalkable(type));
            node_[i][j]->setTileType(TileProperty::getInstance()->getTileType(type));
            if(j+1 < columns_) {
                insertEdge(node_[i][j], node_[i][j+1], 1);
                insertEdge(node_[i][j+1], node_[i][j], 1);

            }
            if(i+1 < rows_) {
                insertEdge(node_[i][j], node_[i+1][j], 1);
                insertEdge(node_[i+1][j], node_[i][j], 1);

            }
            if(i+1 < rows_ && j+1 < columns_) {
                insertEdge(node_[i][j], node_[i+1][j+1], sqrt(2));
                insertEdge(node_[i+1][j+1], node_[i][j], sqrt(2));
            }
            if(i+1 < rows_ && j-1 >= 0) {
                insertEdge(node_[i][j], node_[i+1][j-1], sqrt(2));
                insertEdge(node_[i+1][j-1], node_[i][j], sqrt(2));
            }
        }
    }
}

/**
 * Insere aresta do no' v ao no' w com custo cost
 * @param v No' v
 * @param w No' w
 * @param cost Custo da aresta
 */
void Graph::insertEdge(Node* v, Node* w, float cost) {
    Edge *p;
    if(v == w) return;
    for(p = adj_[v->getX()][v->getY()]; p != NULL; p = p->getNext()) {
        if(p->getNode() == w) return;
    }
    adj_[v->getX()][v->getY()] = new Edge(node_[w->getX()][w->getY()], cost, adj_[v->getX()][v->getY()]);
}

/**
 * Imprime grafo
 */
void Graph::show() {
    Edge *p;
    printf("Nodes: %dx%dzn", rows_, columns_);
    for(int i = 0; i < rows_; i++) {
        for(int j = 0; j < columns_; j++) {
            if(adj_[i][j] != NULL) {
                printf("(%2d,%2d): ", i, j);
                for(p = adj_[i][j]; p != NULL; p = p->getNext()) {
                    printf("(%2d,%2d) ", p->getNode()->getX(), p->getNode()->getY());
                }
                printf("\n");
            }
        }
    }
}

/**
 * Realiza busca do no' s ao no' t usando algoritmo BFS
 * @param s No' s
 * @param t No' t
 * @return true se encontrou caminho, false caso contrario
 */
bool Graph::bfs(Node* s, Node* t) {
    int i, j;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    std::queue<Node*> queue;
    s->setParent(s);
    queue.push(s);
    while(!queue.empty()) {
        current = queue.front();
        queue.pop();
        if(current == t) {
            return true;
        }
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            if(w->isWalkable() && w->getParent() == NULL) {
                w->setParent(current);
                queue.push(w);
            }
        }
    }
    return false;
}

/**
 * Realiza busca do no' s ao no' t usando algoritmo Dijkstra
 * @param s No' s
 * @param t No' t
 * @return true se encontrou caminho, false caso contrario
 */
bool Graph::dijkstra(Node* s, Node* t) {
    int i, j;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setCost(INF);
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    std::priority_queue<Node*, std::vector<Node*>, mycomparison> pq;
    s->setCost(0);
    s->setParent(s);
    pq.push(s);
    while(!pq.empty()) {
        current = pq.top();
        pq.pop();
        if(current == t) {
            return true;
        }
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            if(w->isWalkable()) {
                if(w->getParent() == NULL) {
                    w->setCost(current->getCost() + p->getCost());
                    w->setParent(current);
                    pq.push(w);
                } else {
                    if(w->getCost() > current->getCost() + p->getCost()) {
                        w->setCost(current->getCost() + p->getCost());
                        w->setParent(current);
                    }
                }
            }
        }
    }
    return false;
}

/**
 * Realiza busca do no' s a um vetor de nos' t usando algoritmo Dijkstra
 * @param s No' s
 * @param t Vetor de no's t
 * @return true se encontrou algum no', false caso contrario
 */
bool Graph::groupDijkstra(Node* s, std::vector<Node*> t) {
    int i, j;
    int count = 0;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setCost(INF);
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    std::priority_queue<Node*, std::vector<Node*>, mycomparison> pq;
    s->setCost(0);
    s->setParent(s);
    pq.push(s);
    while(!pq.empty()) {
        current = pq.top();
        pq.pop();
        for(int i = 0; i < (int) t.size(); i++) {
            if(current == t[i]) {
                count++;
            }
        }
        if(count == (int) t.size()) return true;
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            if(w->isWalkable()) {
                if(w->getParent() == NULL) {
                    w->setCost(current->getCost() + p->getCost());
                    w->setParent(current);
                    pq.push(w);
                } else if(w->getCost() > current->getCost() + p->getCost()) {
                    w->setCost(current->getCost() + p->getCost());
                    w->setParent(current);
                }
            }
        }
    }
    return false;
}

/**
 * Busca pelo no' mais proximo de s que possui mesma ZE a alguma do conjunto zones
 * @param s No' s
 * @param zones Conjunto de zonas
 * @return No' encontrado
 * @return true se encontrou algum no', false caso contrario
 */
Node* Graph::noPathDijkstra(Node* s, std::set<int> zones) {
    int i, j;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setCost(INF);
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    //std::auto_ptr<PQueue> pq(new PQueue(M*N+1));
    std::priority_queue<Node*, std::vector<Node*>, mycomparison> pq;
    s->setCost(0);
    s->setParent(s);
    pq.push(s);
    while(!pq.empty()) {
        current = pq.top();
        pq.pop();
        std::set<int>::iterator it;
        for(it = zones.begin(); it != zones.end(); ++it) {
            if((*it) == zoneEquivalency_[current->getZone()]) {
                return current;
            }
        }
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            if(w->getParent() == NULL) {
                w->setCost(current->getCost() + p->getCost());
                w->setParent(current);
                pq.push(w);
            } else if(w->getCost() > current->getCost() + p->getCost()) {
                w->setCost(current->getCost() + p->getCost());
                w->setParent(current);
            }
        }
    }
    return NULL;
}

/**
 * Busca pelo no' mais proximo de s que possui mesma ZE do no' t
 * @param s No' s
 * @param t No' t
 * @return No' encontrado
 */
Node* Graph::noPathDijkstra2(Node* s, Node* t) {
    int i, j;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setCost(INF);
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    //std::auto_ptr<PQueue> pq(new PQueue(M*N+1));
    std::priority_queue<Node*, std::vector<Node*>, mycomparison> pq;
    s->setCost(0);
    s->setParent(s);
    pq.push(s);
    while(!pq.empty()) {
        current = pq.top();
        pq.pop();
        if(zoneEquivalency_[current->getZone()] == zoneEquivalency_[t->getZone()]) {
            return current;
        }
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            if(w->getParent() == NULL) {
                w->setCost(current->getCost() + p->getCost());
                w->setParent(current);
                pq.push(w);
            } else if(w->getCost() > current->getCost() + p->getCost()) {
                w->setCost(current->getCost() + p->getCost());
                w->setParent(current);
            }
        }
    }
    return NULL;
}

/**
 * Busca pelo no' mais proximo de s onde e' possivel construir um edificio do tipo buildType
 * @param s No' s
 * @param buildType Tipo do edificio
 * @return No' encontrado
 */
Node* Graph::dijkstraBuild(Node* s, int buildType) {
    int i, j;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setCost(INF);
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    std::priority_queue<Node*, std::vector<Node*>, mycomparison> pq;
    s->setCost(0);
    s->setParent(s);
    pq.push(s);
    while(!pq.empty()) {
        current = pq.top();
        pq.pop();
        if(isInsertBuildingValid(current, buildType)) {
            return current;
        }
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            if((current->isWalkable() && w->isWalkable()) || !current->isWalkable()) {
                if(w->getParent() == NULL) {
                    w->setCost(current->getCost() + p->getCost());
                    w->setParent(current);
                    pq.push(w);
                } else {
                    if(w->getCost() > current->getCost() + p->getCost()) {
                        w->setCost(current->getCost() + p->getCost());
                        w->setParent(current);
                    }
                }
            }
        }
    }
    return NULL;
}

/**
 * Busca pelo no' mais proximo de s do time inimigo, usado para IA decidir qual ponto atacar
 */
Node* Graph::dijkstraEnemyBuilding(Node* s, int team) {
    int i, j;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setCost(INF);
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    //std::auto_ptr<PQueue> pq(new PQueue(M*N+1));
    std::priority_queue<Node*, std::vector<Node*>, mycomparison> pq;
    s->setCost(0);
    s->setParent(s);
    pq.push(s);
    while(!pq.empty()) {
        current = pq.top();
        pq.pop();
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            // Se o proximo no' for do time inimigo
            if(w->getTeam() != -1 && w->getTeam() != team) {
                return w;
            }
            if(w->isWalkable() || w->getTeam() != -1) {
                if(w->getParent() == NULL) {
                    w->setCost(current->getCost() + p->getCost());
                    w->setParent(current);
                    pq.push(w);
                } else {
                    if(w->getCost() > current->getCost() + p->getCost()) {
                        w->setCost(current->getCost() + p->getCost());
                        w->setParent(current);
                    }
                }
            }
        }
    }
    return NULL;
}

/**
 * Verifica se e' possivel inserir um edificio do tipo buildType no no' s
 * @param s No' s
 * @param buildType Tipo do edificio
 * @return true se for possivel inserir um edificio do tipo buildType no no' s, false caso contrario
 */
bool Graph::isInsertBuildingValid(Node* s, int buildType) {
    int x = s->getX();
    int y = s->getY();
    int dimension = BuildingProperty::getInstance()->getDimension(buildType);
    for(int i = x; i < x+dimension; i++) {
        for(int j = y; j < y+dimension; j++) {
            if(!isInMapDimension(i,j) || BuildingProperty::getInstance()->getTileType(buildType) != node_[i][j]->getTileType()) {
                return false;
            }
        }
    }
    return true;
}

/**
 * Verifica se a posicao (x,y) esta nas dimensoes do mapa
 * @param x Coordenada x
 * @param y Coordenada y
 * @return true se a posicao (x,y) esta nas dimensoes do mapa, false caso contrario
 */
bool Graph::isInMapDimension(int x, int y) {
    if(x >= 0 && y >= 0 && x < rows_ && y < columns_) {
        return true;
    }
    return false;
}

/**
 * Funcao de estimar a distancia entre o no' s ao no' t
 * @return Valor estimado da distancia entre o no' s ao no' t
 */
double Graph::estimate(Node* s, Node* t) {
    return std::sqrt((s->getX()-t->getX())*(s->getX()-t->getX()) + (s->getY()-t->getY())*(s->getY()-t->getY()));
    //return 0;
}

/**
 * Realiza busca do no' s ao no' t usando algoritmo A*
 * @param s No' s
 * @param t No' t
 * @return true se encontrou caminho, false caso contrario
 */
bool Graph::astar(Node* s, Node* t) {
    int i, j;
    Edge* p;
    Node* current;
    Node* w;
    std::vector<std::pair<int,int>> path;
    for(i = 0; i < rows_; i++) {
        for(j = 0; j < columns_; j++) {
            node_[i][j]->setCost(INF);
            node_[i][j]->setParent(NULL);
            node_[i][j]->setOnPath(false);
        }
    }
    s->setDist(0);
    s->setEstimate(estimate(s,t));
    s->setParent(s);
    std::priority_queue<Node*, std::vector<Node*>, mycomparison> pq;
    pq.push(s);
    while(!pq.empty()) {
        current = pq.top();
        pq.pop();
        if(current == t) {
            return true;
        }
        for(p = adj_[current->getX()][current->getY()]; p != NULL; p = p->getNext()) {
            w = p->getNode();
            if(w->isWalkable()) {
                if(w->getParent() == NULL/*w->isOpen() == false*/) {
                    w->setDist(p->getCost() + current->getDist()); // w.dist = custo da aresta(current-w) + current.dist
                    w->setEstimate(estimate(w, t));
                    w->setCost(w->getDist() + w->getEstimate()); // w.cost = w.dist + estimate(w,t)
                    w->setParent(current);
                    pq.push(w);
                } else {
                    if(p->getCost() + current->getDist() < w->getDist()) {
                        w->setDist(p->getCost() + current->getDist()); // w.dist = custo da aresta(current-w) + current.dist
                        w->setParent(current);
                        w->setCost(w->getDist() + w->getEstimate()); // w.cost = w.dist + estimate(w,t)
                    }
                }
            }
        }
    }
    //path.push_back(std::make_pair(t->getX(), t->getY()));
    return false;
}

/**
 * Altera zonas de cada no' do grafo de acordo com a situacao atual
 */
void Graph::setZone() {
    std::set<int> neighbor;
    int currentZone = 1;
    zoneEquivalency_.clear();
    zoneEquivalency_[0] = 0;
    for(int i = 0; i < rows_; i++) {
        for(int j = 0; j < columns_; j++) {
            neighbor.clear();
            if(!node_[i][j]->isWalkable()) {
                node_[i][j]->setZone(0);
                continue;
            }
            // cima-direita
            if(i-1 >= 0 && j+1 < columns_) {
                if(node_[i-1][j+1]->getZone() != 0) {
                    neighbor.insert(node_[i-1][j+1]->getZone());
                }
            }
            // cima
            if (i-1 >= 0) {
                if(node_[i-1][j]->getZone() != 0) {
                    neighbor.insert(node_[i-1][j]->getZone());
                }
            }
            // cima-esquerda
            if(i-1 >= 0 && j-1 >= 0) {
                if(node_[i-1][j-1]->getZone() != 0) {
                    neighbor.insert(node_[i-1][j-1]->getZone());
                }
            }
            // equerda
            if(j-1 >= 0) {
                if(node_[i][j-1]->getZone() != 0) {
                    neighbor.insert(node_[i][j-1]->getZone());
                }
            }
            // Se todos vizinhos anteriores não foram populados
            if(neighbor.size() == 0) {
                node_[i][j]->setZone(currentZone);
                zoneEquivalency_[currentZone] = currentZone;
                currentZone++;
            } else {
                std::set<int>::iterator it;
                for(it = neighbor.begin(); it != neighbor.end(); ++it) {
                    zoneEquivalency_[(*it)] = zoneEquivalency_[(*neighbor.begin())];
                }
                node_[i][j]->setZone((*neighbor.begin()));
            }
        }
    }
    // Aprofunda a zona de equivalencia ate que nao seja mais possivel
    for(int i = 0; i < (int) zoneEquivalency_.size(); i++) {
        int key = i;
        while(zoneEquivalency_[key] != key) {
            key = zoneEquivalency_[key];
        }
        zoneEquivalency_[i] = zoneEquivalency_[key];
    }
}

/**
 * Altera zonas de cada no' do grafo de acordo com a situacao atual, outro modo
 */
void Graph::setZone2() {

    int currentZone = 1;
    zoneEquivalency_.clear();
    zoneEquivalency_[0] = 0;
    for(int i = 0; i < rows_; i++) {
        for(int j = 0; j < columns_; j++) {
            node_[i][j]->setZone(-1);
        }
    }

    for(int i = 0; i < rows_; i++) {
        for(int j = 0; j < columns_; j++) {

            if(!node_[i][j]->isWalkable()) {
                node_[i][j]->setZone(0);
                continue;
            }

            if (node_[i][j]->getZone() < 0) {
                node_[i][j]->setZone(currentZone);
                zoneEquivalency_[currentZone] = currentZone;
                currentZone++;
            }

            // direita
            if (j+1 < columns_) {
                int X=i;
                int Y=j+1;

                if (node_[X][Y]->isWalkable()) {
                    if (node_[X][Y]->getZone() < 0) {
                        node_[X][Y]->setZone(node_[i][j]->getZone());
                    } else {
                        if (zoneEquivalency_[node_[X][Y]->getZone()] < zoneEquivalency_[node_[i][j]->getZone()]) {
                            zoneEquivalency_[node_[i][j]->getZone()] = zoneEquivalency_[node_[X][Y]->getZone()];
                        } else {
                            zoneEquivalency_[node_[X][Y]->getZone()] = zoneEquivalency_[node_[i][j]->getZone()];
                        }
                    }
                }
            }

            // baixo-direita
            if (i+1 < rows_ && j+1 < columns_) {
                int X=i+1;
                int Y=j+1;

                if (node_[X][Y]->isWalkable()) {
                    if (node_[X][Y]->getZone() < 0) {
                        node_[X][Y]->setZone(node_[i][j]->getZone());
                    } else {
                        if (zoneEquivalency_[node_[X][Y]->getZone()] < zoneEquivalency_[node_[i][j]->getZone()]) {
                            zoneEquivalency_[node_[i][j]->getZone()] = zoneEquivalency_[node_[X][Y]->getZone()];
                        } else {
                            zoneEquivalency_[node_[X][Y]->getZone()] = zoneEquivalency_[node_[i][j]->getZone()];
                        }
                    }
                }
            }

            // baixo
            if (i+1 < rows_) {
                int X=i+1;
                int Y=j;

                if (node_[X][Y]->isWalkable()) {
                    if (node_[X][Y]->getZone() < 0) {
                        node_[X][Y]->setZone(node_[i][j]->getZone());
                    } else {
                        if (zoneEquivalency_[node_[X][Y]->getZone()] < zoneEquivalency_[node_[i][j]->getZone()]) {
                            zoneEquivalency_[node_[i][j]->getZone()] = zoneEquivalency_[node_[X][Y]->getZone()];
                        } else {
                            zoneEquivalency_[node_[X][Y]->getZone()] = zoneEquivalency_[node_[i][j]->getZone()];
                        }
                    }
                }
            }

            // baixo-esquerda
            if (i+1 < rows_ && j-1 >= 0) {
                int X=i+1;
                int Y=j-1;

                if (node_[X][Y]->isWalkable()) {
                    if (node_[X][Y]->getZone() < 0) {
                        node_[X][Y]->setZone(node_[i][j]->getZone());
                    } else {
                        if (zoneEquivalency_[node_[X][Y]->getZone()] < zoneEquivalency_[node_[i][j]->getZone()]) {
                            zoneEquivalency_[node_[i][j]->getZone()] = zoneEquivalency_[node_[X][Y]->getZone()];
                        } else {
                            zoneEquivalency_[node_[X][Y]->getZone()] = zoneEquivalency_[node_[i][j]->getZone()];
                        }
                    }
                }
            }
        }
    }

    //printf("rows: %d, columns: %d\n", rows_, columns_);
    /*
    for(int i = 0; i < (int) zoneEquivalency_.size(); i++) {
        printf("ZE[%d] = %d\n", i, zoneEquivalency_[i]);
    }
    printf("\n");
    */
    /*
    std::map<int,int>::iterator it;
    for(it = zoneEquivalency_.begin(); it != zoneEquivalency_.end(); ++it) {
        printf("ZE[%d] = %d\n", (*it).first, (*it).second);
    }
    */
}
